import numpy as np
import math
import os, sys
import pygame

def getModpath():
    
    np_path = pygame.__file__
    if os.path.isfile(np_path[:-19] + r'\astron\__init__.py'):
        return np_path[:-19] + r'\astron'
    else:
        modpath = os.path.abspath(os.path.split(sys.argv[0])[0])        
        return modpath

class Velocity:
    
    def __init__(self, x_vel, y_vel):
        
        self.x = x_vel
        self.y = y_vel
        self.vec = [self.x, self.y]
        self.rot_matrix = getRotMatrix(self.getTheta())
        self.mag = (self.x ** 2 + self.y ** 2) ** 0.5
        self.theta = self.getTheta()
    
    def getTheta(self):
        
        angle = angleBetween([1,0], self.vec)
        if self.y < 0:
            angle += np.pi
        
        return angle
    
    def __repr__(self):
        return str((self.x, self.y))

class Force:
    
    def __init__(self, x_vector, y_vector, mag):
        
        hyp = (x_vector ** 2 + y_vector ** 2) ** 0.5
        if hyp != 0.0:
            ratio = mag / hyp
        else:
            ratio = 0.0
            mag = 0.0
            
        self.x = x_vector * ratio
        self.y = y_vector * ratio
        self.mag = mag  
    
    def __repr__(self):  
         
        return str((self.x, self.y))      

class Momentum:
    
    def __init__(self, x_vel, y_vel, mass = 1):
        
        self.x = mass * x_vel
        self.y = mass * y_vel
    
    @classmethod
    def fromImpulse(cls, force = Force, duration = float):
        
        x = force.x * duration
        y = force.y * duration
        
        return cls(x, y)

    def __add__(self, new):
        
        self.x += new.x
        self.y += new.y
        
        return self 

    def __repr__(self):
        
        return str((self.x, self.y))
    
class Orbit:
    
    def __init__(self, a, b, center_x, center_y, progress = 0.0, CW = True, angular_step = 3.14/900):
        self.a = a
        self.b = b
        self.center_x = center_x
        self.center_y = center_y
        self.progress = progress
        self.cw = CW
        self.angular_step = angular_step % 2*np.pi 
        
    def x(self, progress):
        return self.a * np.cos(progress) + self.center_x
    
    def y(self, progress):
        return self.b * np.sin(progress) + self.center_y
    
    def nextPos(self, factor = 1.0):
        if self.cw:
            self.progress += self.angular_step * factor
        else:
            self.progress -= self.angular_step * factor
            
        return self.x(self.progress), self.y(self.progress)

    def resetPos(self):
        self.progress = 0
        return self.x(self.progress), self.y(self.progress)
 
def unit_vector(vector):
    
    """ Returns the unit vector of the vector.  """
    if np.linalg.norm(vector) > 0:
        return vector / np.linalg.norm(vector)
    else:
        return np.array([0.0, 0.0])
        
def getRotMatrix(theta):
    
    return [
        [math.cos(theta), -math.sin(theta)],
        [math.sin(theta), math.cos(theta)]
    ]

def angleBetween(v1, v2):
    """ Returns the angle in degrees between vectors 'v1' and 'v2'::

            >>> angle_between((1, 0, 0), (0, 1, 0))
            1.5707963267948966
            >>> angle_between((1, 0, 0), (1, 0, 0))
            0.0
            >>> angle_between((1, 0, 0), (-1, 0, 0))
            3.141592653589793
    """
    v1_u = unit_vector(v1)
    v2_u = unit_vector(v2)
    return np.arccos(np.clip(np.dot(v1_u, v2_u), -1.0, 1.0))


import sys, os
import numpy as np
import pygame
import math
from .utilities import *

DEFAULT_SC_SPRITE = getModpath() + r'\images\ship1.png'

class Asset:
    
    def __init__(self, name, x = 0.0, y = 0.0, mass = 0, vel = None):
        
        self.x = x
        self.y = y
        self.name = name
        self.mass = mass
        self.vel = vel
        if not vel:
            self.vel = Velocity(0.0, 0.0)
            
        self._p = Momentum(self.vel.x, self.vel.y, self.mass)
    
    def resetPos(self):
        
        self.x = 0
        self.y = 0
    
    def calcDistance(self, other_asset):
        
        dx = self.x - other_asset.x
        dy = self.y - other_asset.y
        
        return np.sqrt(dx**2 + dy**2)   

    def calcVector(self, other_asset):
        
        dx = other_asset.x - self.x
        dy = other_asset.y - self.y
        
        return dx, dy
    
    def calcGravitationalForce(self, other_asset):
       
        G = 6.67408e-11
        M = self.mass
        m = other_asset.mass
        r = self.calcDistance(other_asset)
        
        mag = G*M*m / r**2
        x,y = self.calcVector(other_asset)
        
        return Force(x, y, mag)
     
    @property
    def p(self):
        return self._p

    @p.setter
    def p(self, val):
        self._p = val
        self.vel = Velocity(val.x / self.mass, val.y / self.mass)
              
class Planet(Asset):
    
    def __init__(self, name, mass = 0.0, orbit = Orbit, color = (100,100,100)):
    
        super().__init__(name, 0.0, 0.0, mass)
        self.g = 6.67408e-11 # m^3/kg*s^2
        self.orbit = orbit
        self.color = color
        self.move()
            
    def move(self, dt = 1.0):
        
        self.x, self.y = self.orbit.nextPos(dt)

class Sprite:
    
    def __init__(self, image_path = DEFAULT_SC_SPRITE, size = (100, 100), theta_deg_offset = 90.0):
    
        self.size = size
        self.image_path = image_path
        self.theta_deg_offset = theta_deg_offset
        
        self.checkPath(image_path)
        self._loadAllImages(image_path)
        
    def checkPath(self, image_path):
        
        if image_path[-3:] != 'png':
            raise ValueError('Must be a PNG image.')
        
    def _loadAllImages(self, image_path):
        
        self.base = pygame.image.load(image_path)
        self.px = pygame.image.load(image_path[:-4] + '_+x' + '.png')
        self.py = pygame.image.load(image_path[:-4] + '_+y' + '.png')
        self.nx = pygame.image.load(image_path[:-4] + '_-x' + '.png')
        self.ny = pygame.image.load(image_path[:-4] + '_-y' + '.png')
    
    def transform(self, x, y, radians, thruster_direction = None):
        
        ''' Return PyGame Image + Rectangle objects for scree.blit onto screen. '''
        
        img = self.loadThrusterImage(thruster_direction)
        sc_rot = pygame.transform.rotate(img, math.degrees(radians) + self.theta_deg_offset)
        sc_rect = sc_rot.get_rect()
        sc_rect = sc_rect.move((x-sc_rect.centerx, y-sc_rect.centery))
        
        self.sprite = sc_rot
        self.rect = sc_rect 
        
        return sc_rot, sc_rect
    
    def loadThrusterImage(self, thrust_direction = None):
        
        if thrust_direction:
            if thrust_direction=='+x':
                img = pygame.transform.scale(self.px, self.size)
            if thrust_direction=='+y':
                img = pygame.transform.scale(self.py, self.size)
            if thrust_direction=='-x':
                img = pygame.transform.scale(self.nx, self.size)
            if thrust_direction=='-y':
                img = pygame.transform.scale(self.ny, self.size)
        
        else:
            img = pygame.transform.scale(self.base, self.size)
        
        return img
    
    # def _render(self, x,y, theta_degrees, thrust_direction = None):
    
    #     sc_rot, sc_rect = self.transform(x, y, theta_degrees)
        
class Spacecraft(Asset):
    
    def __init__(self, name, mass = 0.0, gas_level = 0.0, thrust_force = 0.0, sprite = None):
        
        super().__init__(name, 0.0, 0.0, mass)
        self.gas_level = gas_level
        self._initial_gas_level = gas_level
        self.thrust = False
        self.thrust_direction = '-y' # +/-x,-y
        self.thrust_mag = thrust_force
        self.sprite = sprite
        
        # Load default if none given
        if not sprite:
            self.sprite = Sprite()
    
    def bodyTransform(self, vector):
        
        ''' Body pointing towards self.vel '''
        
        return np.matmul(self.vel.rot_matrix, vector)
       
    def getThrustImpulse(self, time):
        
        if self.gas_level <= 0.0:
            self.gas_level = 0.0
            self.thrust = False
        
        if self.thrust:
            
            self.gas_level -= self.thrust_mag / 1000
            
            vel_vec = self.vel.vec
            if np.linalg.norm(self.vel.vec) == 0.0:
                vel_vec = [0,-1]                
            
            if self.thrust_direction == '-y':
                # vec = [0,-1]
                vector = vel_vec
            elif self.thrust_direction == '+y':
                # vec = [0,1]
                vector = np.matmul(getRotMatrix(np.pi), vel_vec)
            elif self.thrust_direction == '-x':
                # vec = [1,0]
                vector = np.matmul(getRotMatrix(np.pi/2), vel_vec)
            elif self.thrust_direction == '+x':
                # vec = [-1,0]
                vector = np.matmul(getRotMatrix(np.pi * 1.5), vel_vec)
                
            force = Force(vector[0], vector[1], self.thrust_mag)
            # print(math.degrees(angleBetween(self.vel.vec, [force.x,force.y])))  
            return Momentum.fromImpulse(force, time)
        
        return Momentum(0.0, 0.0)
    
    def setNetMomentum(self, impulse_time, external_force = None):
        
        # Thrust impulse
        thrust_i = self.getThrustImpulse(impulse_time)
        
        # External impulse
        if external_force:
            external_i = Momentum.fromImpulse(external_force, impulse_time) 
        else:
            external_i = Momentum(0.0, 0.0)
        
        self.p = self.p + thrust_i + external_i       
    
    def move(self, time):
        
        self.x += self.vel.x * time
        self.y += self.vel.y * time
    
    def reset(self, sc_start_pos = None):
        
        if sc_start_pos:
            self.x, self.y = sc_start_pos
        self.p = Momentum(0.0, 0.0)
        self.thrust = False
        self.gas_level = self._initial_gas_level
               
    @property
    def p(self):
        return self._p

    @p.setter
    def p(self, val):
        self._p = val
        self.vel = Velocity(val.x / self.mass, val.y / self.mass)
        if self.thrust:
            self.sprite.transform(self.x, self.y, self.vel.theta, self.thrust_direction)
        else:
            self.sprite.transform(self.x, self.y, self.vel.theta)        
       
        



import pygame
import os, sys
import math
import numpy as np
import time
import ctypes

from .assests import *
from .scene import *
from .utilities import *

class GameScene(Scenario):
    
    def __init__(self, 
                 resolution,
                 sc,
                 planets,
                 sc_start_pos = None, 
                 win_region = tuple, # ([x1,x2], [y1,y2])
                 win_velocity = 0.0,
                 win_region_color = (0.0, 255, 174),
                 background = None # Image path / color tuple
                ):
        
        super().__init__(resolution, sc, planets, sc_start_pos)
        self.background = background
        self.win_region = win_region
        self.win_min_velocity = win_velocity
        self.win_region_color = win_region_color
        self._attempts = 0
        
class Game:
    
    def __init__(self, fullscreen = False, font = 'Helvetica Neue', font_size = 30, fps = 60.0, scenes = []):
        
        pygame.init()
        pygame.font.init()
        
        self.clock = pygame.time.Clock()
        self.fullscreen = fullscreen
        self.fps = fps
        self.font = pygame.font.SysFont(font, font_size)
        self.scenes = scenes
        self.screen = None # current screen
        self.current_scene = self.scenes[0]
        self._bg_loaded = False
        self.__rpk = 45 / 4e16 # radius per kilogram of planet
    
    def loadImage(self, path, size):
        
        img = pygame.image.load(path)
        img_scaled = pygame.transform.scale(img, size)
        rectangle = img_scaled.get_rect()
        rectangle = rectangle.move((0,0))
        
        return img_scaled, rectangle
    
    def renderBackground(self, scene):
        
        ''' Render background of scene on current game screen '''
              
        if scene.background:
            if (isinstance(scene.background, list) or isinstance(scene.background, tuple)) and len(scene.background) == 3:
                # RGB Value given
                self.screen.fill(scene.background)
            else:
                # Image path given
                if not self._bg_loaded:
                    self._bg_img, self._bg_rect = self.loadImage(scene.background, scene.size)                  
                
                self.screen.blit(self._bg_img, self._bg_rect)   
                  
        else:
            # Default: lack screen
            self.screen.fill((0,0,0))
    
    def renderPlanets(self, scene):
        
        rpk = self.__rpk
        
        for planet in scene.planets:
            
            pygame.draw.ellipse(self.screen, planet.color, pygame.Rect(planet.x-rpk*planet.mass, planet.y-rpk*planet.mass, rpk*planet.mass * 2, rpk*planet.mass * 2))
            # Orbit
            pygame.draw.ellipse(self.screen, (255,255,255), pygame.Rect(planet.orbit.center_x-planet.orbit.a, planet.orbit.center_y-planet.orbit.b, planet.orbit.a*2, planet.orbit.b*2), 1)
    
    def renderWinRegion(self, scene):
        
        pygame.draw.line(self.screen, scene.win_region_color, (scene.win_region[0][0], scene.win_region[0][1]), (scene.win_region[1][0], scene.win_region[1][1]), 25)
    
    def createScreen(self, scene = None):
        
        ''' Start screen window '''
        
        if not scene:
            scene = self.current_scene
        
        if self.fullscreen:
            
            infoObject = pygame.display.Info()
            screen_x, screen_y = infoObject.current_w, infoObject.current_h
            
            if scene.size[0] == screen_x or scene.size[1] == screen_y:
                # Only windows!            
                ctypes.windll.user32.SetProcessDPIAware()
            
            self.screen = pygame.display.set_mode((scene.size[0], scene.size[1]), pygame.FULLSCREEN)
            
        else:
            self.screen = pygame.display.set_mode((scene.size[0], scene.size[1]))
    
    def renderSc(self, scene):
        
        sc_rot = self.current_scene.sc.sprite.sprite
        sc_rect = self.current_scene.sc.sprite.rect
        
        self.screen.blit(sc_rot, sc_rect)
    
    def renderScene(self, scene):
        
        '''
        Render scene onto game screen
        '''
        
        if pygame.display.get_surface().get_size() != scene.size:
            self.createScreen(scene)
            
        # background
        self.renderBackground(scene)
        
        # planets
        self.renderPlanets(scene)
        
        # win region
        self.renderWinRegion(scene)
        
        # spacecraft
        self.renderSc(scene)
        
        # Hu
        self.renderHud(scene)
    
    def renderHud(self, scene):
        
        sc = self.current_scene.sc
        
        # Gas level
        text_surface = self.font.render('Gas: ' + str(sc.gas_level), True, (255,255,255))
        self.screen.blit(text_surface, (scene.size[0]-110, scene.size[1]-60))
        
        # Velocity
        text_surface = self.font.render('Velocity: ' + str(round(sc.vel.mag,2)), True, (255,255,255))
        self.screen.blit( text_surface, (scene.size[0]-145, scene.size[1]-20))
        
        # Escape velocity needed
        text_surface = self.font.render('Velocity Required: ' + str(round(self.current_scene.win_min_velocity, 2)), True, (255,255,255))
        self.screen.blit(text_surface, (0.0, scene.size[1]-20))
        
        # Attempts
        text_surface = self.font.render('Attempts: ' + str(self.current_scene._attempts), True, (255,255,255))
        self.screen.blit(text_surface, (scene.size[0]-150, 10.0))
        
        # Level counter
        text_surface = self.font.render('Level ' + str(self.scenes.index(self.current_scene)+1), True, (255,255,255))
        self.screen.blit(text_surface, (scene.size[0]-150, 30))
    
    def captureSpacecraftControls(self, event):
            
        if event.type == pygame.KEYDOWN and event.key in [pygame.K_DOWN, pygame.K_UP, pygame.K_LEFT, pygame.K_RIGHT]:
            
            self.current_scene.sc.thrust = True
            if event.key == pygame.K_DOWN:
                    self.current_scene.sc.thrust_direction = '+y'
            if event.key == pygame.K_UP:
                    self.current_scene.sc.thrust_direction = '-y'
            if event.key == pygame.K_RIGHT:
                    self.current_scene.sc.thrust_direction = '-x'
            if event.key == pygame.K_LEFT:
                    self.current_scene.sc.thrust_direction = '+x'
        elif event.type == pygame.KEYUP and event.key in [pygame.K_DOWN, pygame.K_UP, pygame.K_LEFT, pygame.K_RIGHT]:
        
            self.current_scene.sc.thrust = False                
    
    def checkSceneWin(self, scene):
        
        sc = self.current_scene.sc
        screen_x = self.current_scene.size[0]
        screen_y = self.current_scene.size[0]
        win_region_1 = self.current_scene.win_region[0]
        win_region_2 = self.current_scene.win_region[1]
        
        won = False
        failed = False
        
        # Win if
        walls = [ 0, screen_x, screen_y]
        for wall in walls:
            if not won:
                if win_region_1[0] == wall and win_region_2[0] == wall:
                    if sc.x <= wall  and win_region_1[1] <= sc.y <= win_region_2[1] and sc.vel.mag >= self.current_scene.win_min_velocity:
                        won = True
                elif win_region_1[1] == wall and win_region_2[1] == wall:
                    if sc.y <= wall  and win_region_1[0] <= sc.x <= win_region_2[0] and sc.vel.mag >= self.current_scene.win_min_velocity:
                        won = True 
        
        if win_region_1[0] <= sc.x <= win_region_2[0]  and win_region_1[1] <= sc.y <= win_region_2[1] and sc.vel.mag >= self.current_scene.win_min_velocity:
            won = True
                
        # Out of bounds
        if not 0.0 < sc.x < self.current_scene.size[0] or not 0.0 < sc.y < self.current_scene.size[1]:
            failed = True
        
        # Collisions
        for planet in self.current_scene.planets:
            planet_r = self.__rpk*planet.mass
            if planet.x - planet_r <= sc.x <= planet.x + planet_r and planet.y - planet_r <= sc.y <= planet.y + planet_r:
                failed = True
        
        return won, failed
    
    def renderFullscreenDialog(self, texts, xoffsets = [], yoffsets = [], sleep_time = 1.5):
        
        screen_x = self.current_scene.size[0]
        screen_y = self.current_scene.size[1]
        self.screen.fill((0, 0, 0))
        c = 0
        
        if not xoffsets:
            [xoffsets.append(0) for i in range(len(texts))]
        if not yoffsets:
            [yoffsets.append(0) for i in range(len(texts))]
        
        for text in texts:
            text_surface = self.font.render(text, True, (255,255,255))
            self.screen.blit( text_surface, (screen_x/2 + xoffsets[c], screen_y/2 + yoffsets[c])) 
            c+=1 
            
        pygame.display.update()
        
        time.sleep(sleep_time)
        
    def nextScene(self, done):
        
        current_i = self.scenes.index(self.current_scene)
        
        if current_i < len(self.scenes) - 1:
            self.current_scene  = self.scenes[current_i+1]
            return self.scenes[current_i+1], done
        else:
            return self.current_scene, True
    
    def renderFinalMessage(self, won, failed):
        
        center_x = self.current_scene.size[0] / 2
        
        if won:
            
            self.renderFullscreenDialog([
                'You\'re a gravity assist pro :D',
                'Final score: ' + str(self.calcScore())
                ], 
                xoffsets = [-100, -50.0], yoffsets = [0, 100])
        
        else:
            
            self.renderFullscreenDialog(['No worries, see ya next time!', 
                                         'Final score: ' + str(self.calcScore())], xoffsets = [-100, -50.0], yoffsets = [0, 100])
        
    def calcScore(self):
        
        per_scene_score = 100.0
        attempt_deductions = 10.0
        
        total = 0.0
        
        for scene in self.scenes:
            if scene._attempts > 1:
                total += per_scene_score - (scene._attempts-1) * attempt_deductions
            
        return total            
    
    def startGame(self, scene_to_start_at = None, splash = True):
        
        self.createScreen(scene_to_start_at)
        if splash:
            self.renderFullscreenDialog([
                'ASTRON', 
                'Use arrow keys to get to the green region with the required velocity!'
                                        
            ], xoffsets = [-50, -300.0], yoffsets = [0, 100], sleep_time= 5)
        
        dt = 1 / self.fps
        done = False        
        while not done:
            
            # check game exit conditions
            for event in pygame.event.get():
                if event.type == pygame.KEYDOWN and event.key == pygame.K_ESCAPE:
                        done = True
                    
                # Modify spacecraft thrusters 
                self.captureSpacecraftControls(event)
                        
            # Iterate next planetary + sc positions
            self.current_scene.updateAllPos(dt)
                
            # Check scene exit conditions
            won, failed = self.checkSceneWin(self.current_scene)
            if won:
                self.renderFullscreenDialog(['Won!'], xoffsets=[-10])
                self.current_scene, done = self.nextScene(done)
                self.current_scene._attempts += 1
            elif failed:
                self.renderFullscreenDialog(['Oops, try again!'], xoffsets=[-75])
                self.current_scene.resetPos()
                self.current_scene._attempts += 1
    
            if not done:
                # Draw modified scene            
                self.renderScene(self.current_scene)
                pygame.display.update()
                dt = self.clock.tick(self.fps) / 1000

        self.renderFinalMessage(won, failed)
        pygame.quit()
        
          
        


import os,sys
from setuptools import setup

# For regular package imports, not 'import .packagename'
# modpath = os.path.abspath(os.path.split(sys.argv[0])[0]) + '/astron'
# sys.path.append(modpath)
# import astron

setup(name='astron',
      version='0.1.8',
      description='...',
      url='https://www.python.org/sigs/distutils-sig/',
      author='Addy Bhatia',
      author_email='jude.addy999@gmail.com',
      license='MIT',
      install_requires=['autopep8', 'pygame', 'numpy'],
      packages=['astron'], 
      include_package_data=True,
      zip_safe=True)
